﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace com.tomfosdick.ThreadingExamples
{
    class Program
    {
        #region Song to the Men of England - Percy Bysshe Shelley
        static string[] Shelley =
        {
            "Men of England, wherefore plough",
            "For the lords who lay ye low?",
            "Wherefore weave with toil and care",
            "The rich robes your tyrants wear?",

            "Wherefore feed and clothe and save,",
            "From the cradle to the grave,",
            "Those ungrateful drones who would",
            "Drain your sweat -nay, drink your blood?",

            "Wherefore, Bees of England, forge",
            "Many a weapon, chain, and scourge,",
            "That these stingless drones may spoil",
            "The forced produce of your toil?",

            "Have ye leisure, comfort, calm,",
            "Shelter, food, love's gentle balm?",
            "Or what is it ye buy so dear",
            "With your pain and with your fear?",

            "The seed ye sow another reaps;",
            "The wealth ye find another keeps;",
            "The robes ye weave another wears;",
            "The arms ye forge another bears.",

            "Sow seed, -but let no tyrant reap;",
            "Find wealth, -let no imposter heap;",
            "Weave robes, -let not the idle wear;",
            "Forge arms, in your defence to bear.",

            "Shrink to your cellars, holes, and cells;",
            "In halls ye deck another dwells.",
            "Why shake the chains ye wrought? Ye see",
            "The steel ye tempered glance on ye.",

            "With plough and spade and hoe and loom,",
            "Trace your grave, and build your tomb,",
            "And weave your winding-sheet, till fair",
            "England be your sepulchre!",
        };
        #endregion

        static Semaphore mySemaphore = new Semaphore(0, int.MaxValue);
        static Queue<string> myQueue = new Queue<string>();

        static Thread[] producers = new Thread[2];
        static Thread[] consumers = new Thread[3];

        static Random random = new Random();

        static void Main(string[] args)
        {
            for (int i = 0; i < producers.Length; i++)
            {
                producers[i] = new Thread(new ThreadStart(Producer));
                producers[i].Start();
            } 

            //Wait so that there's some in the queue when the 
            //consumers come online
            Thread.Sleep(5000 + random.Next(5000));
            for (int i = 0; i < consumers.Length; i++)
            {
                consumers[i] = new Thread(new ThreadStart(Consumer));
                consumers[i].Start();
            }

            #region cleanup
            //clean up
            foreach (Thread t in producers) { t.Join(); }
            lock (myQueue)
            {
                for (int i = 0; i < consumers.Length; i++)
                {
                    myQueue.Enqueue(null);
                    mySemaphore.Release();
                }
            }
            foreach(Thread t in consumers){ t.Join();}
            #endregion
            Console.Write("Press any key to exit...");
            Console.ReadKey();
        }

        static void Consumer()
        {
            string item;
            int queueLength;
            do
            {
                mySemaphore.WaitOne(); // wait for an item to be available
                lock (myQueue) // queue isn't thread safe so we need to lock.
                {
                    item = myQueue.Dequeue();
                    queueLength = myQueue.Count;
                }
                if (!String.IsNullOrEmpty(item))
                    Console.WriteLine(String.Format("C{0:00} DEQUEUE Q{1:00}: {2}", 
                        Thread.CurrentThread.ManagedThreadId, queueLength, item));
                Thread.Sleep(250);
            } while (!String.IsNullOrEmpty(item)); //we use null to signify end of processing
        }

        static void Producer()
        {
            int queueLength;
            foreach (string line in Shelley)
            {
                Thread.Sleep(random.Next(1000));
                lock(myQueue) //queue isn't thread safe so we need to lock.
                {
                    myQueue.Enqueue(line);
                    queueLength = myQueue.Count;
                }
                mySemaphore.Release(); //let any waiting consumers go.
                Console.WriteLine(String.Format("p{0:00} enqueue Q{1:00}: {2} ",
                    Thread.CurrentThread.ManagedThreadId, queueLength, line));
            }
        }
    }
}
